"""
================================
dms_variants
================================

Python package for analyzing deep mutational scanning of barcoded
variants of genes.

"""

__author__ = '`the Bloom lab <https://research.fhcrc.org/bloom/en.html>`_'
__email__ = 'jbloom@fredhutch.org'
__version__ = '0.7.1'
__url__ = 'https://github.com/jbloomlab/dms_variants'
